#pragma once
#include "glwindows.h"

#include "tricks.h"
#ifndef SWIG
#include <vector>
#endif

namespace glc 
{

#define ASSERT_BUFFER_BOUND_IF_TARGET(TARGET) assert(BufferType!=TARGET || glGetIntegerValue(TARGET##_BINDING) == BufferName);
#define ASSERT_BUFFER_BOUND(NAME, TARGET) assert(glGetIntegerValue(TARGET##_BINDING) == NAME)

  GLint glGetIntegerValue(GLenum pname);

  class CommonBasicBuffer;
  TYPEDEF_AUTO_PTR_TYPE(CommonBasicBuffer);

  class CommonBasicBuffer
  {
  protected:
    GLuint BufferName;
    GLuint BufferType;
  public:
    friend class CommonVertexAttribBufferDecorator;
    CommonBasicBuffer(GLuint BufferType);
    virtual ~CommonBasicBuffer();
    operator GLuint () {return BufferName;}
    void * BindMap(GLintptr offset, GLsizeiptr size);
    void Unmap();
    void Bind();
    void Unbind();
    void SetData(const GLvoid* ptr, GLsizeiptr size, GLenum hint = GL_STATIC_DRAW);
  };

  template <typename T>
  class CommonTypedBuffer : public CommonBasicBuffer
  {
  public:
    CommonTypedBuffer(GLuint buffer_type, const std::vector<T>& data);
  };
  #ifdef SWIG
  %template (CommonTypedBuffer_GLuint) CommonTypedBuffer< GLuint >;
	#endif

  class CommonElementArrayBuffer :public  CommonTypedBuffer<GLuint>
  {
  public:
    static const GLenum INDEX_TYPE = GL_UNSIGNED_INT;
    GLenum PrimitiveType;
    GLsizei IndicesCount;
    CommonElementArrayBuffer(std::vector<GLuint>& data, GLenum primType);
    void DrawInstanced(GLuint instanceCount=1); 
  };

  class CommonVertexAttribute;
  struct AttributeSpec
  {
    size_t Size;
    GLenum Type;
    GLsizei Stride;
    size_t Offset;
    AttributeSpec(size_t size, GLenum type, GLsizei stride, size_t off =0)
      : Type(type), Stride(stride), Size(size), Offset(off) {} 
  };

  static inline AttributeSpec make_float_attribute(GLsizei number_of_components)
  { 
    return AttributeSpec(number_of_components, GL_FLOAT, number_of_components*sizeof(float));
  }

  template<typename T>
  std::shared_ptr< CommonTypedBuffer<T> > make_vertex_buffer(
    const std::vector<T> &data)
  {
    return make_auto_ptr(
      new CommonTypedBuffer<T>(GL_ARRAY_BUFFER, data));
  }



  template <typename T>
  class CommonUniformBuffer : public CommonBasicBuffer
  {
    GLint UniformBufferOffset;
  public:
    CommonUniformBuffer( );
    void UploadUniform(T& value);
    void BindBase(GLuint index);
  };
#pragma endregion buffer objects

#pragma region buffer implementation
  template <typename T>
  CommonUniformBuffer<T>::CommonUniformBuffer():
  CommonBasicBuffer(GL_UNIFORM_BUFFER), UniformBufferOffset(0)
  {
    //printf("%s", "created uniform buffer");
    ASSERT_BUFFER_BOUND(BufferName, GL_UNIFORM_BUFFER);
    UniformBufferOffset = glGetIntegerValue(GL_UNIFORM_BUFFER_OFFSET_ALIGNMENT);
    Unbind();
    GLint UniformBlockSize = std::max(GLint(sizeof(T)), UniformBufferOffset);
    SetData(NULL, UniformBlockSize,  GL_DYNAMIC_DRAW);
     
  }

  template <typename T>
  void CommonUniformBuffer<T>::BindBase( GLuint index )
  {
    glBindBufferBase(BufferType, index, BufferName);
  }

  template <typename T>
  void CommonUniformBuffer<T>::UploadUniform(T& value)
  {  
    //printf("%s", "maping uniform buffer");
    glm::mat4* Pointer = (T*)BindMap(0,	sizeof(T));
    *Pointer = value;
    Unmap();
    //printf("%s", "unmpaped uniform buffer");
  }

  template<typename T>
  inline  CommonTypedBuffer<T>* make_buffer(GLuint buffer_type, std::vector<T>& data)
  {
    return new CommonTypedBuffer<T>(buffer_type, data);
  }

  template <typename T>
  CommonTypedBuffer<T>::CommonTypedBuffer(GLuint buffer_type, const std::vector<T>& data)
    :CommonBasicBuffer(buffer_type)
  {
    SetData(&data[0], sizeof(T)*data.size());
  }

 
};
 