
#include "error.hpp"
#include "compiler.hpp"
#include "glo_external.h"

namespace glf
{
  bool initDebugOutput();
  bool checkGLVersion(GLint MajorVersionRequire, GLint MinorVersionRequire);
  bool checkExtension(char const * String);

  inline int version(int Major, int Minor)
  {
    return Major * 100 + Minor * 10;
  }
  inline bool checkError(const char* Title)
  {
    int Error;
    if((Error = glGetError()) != GL_NO_ERROR)
    {
      std::string ErrorString;
      switch(Error)
      {
      case GL_INVALID_ENUM:
        ErrorString = "GL_INVALID_ENUM";
        break;
      case GL_INVALID_VALUE:
        ErrorString = "GL_INVALID_VALUE";
        break;
      case GL_INVALID_OPERATION:
        ErrorString = "GL_INVALID_OPERATION";
        break;
      case GL_INVALID_FRAMEBUFFER_OPERATION:
        ErrorString = "GL_INVALID_FRAMEBUFFER_OPERATION";
        break;
      case GL_OUT_OF_MEMORY:
        ErrorString = "GL_OUT_OF_MEMORY";
        break;
      default:
        ErrorString = "UNKNOWN";
        break;
      }
      fprintf(stdout, "OpenGL Error(%s): %s\n", ErrorString.c_str(), Title);
    }
    return Error == GL_NO_ERROR;
  }

  bool checkFramebuffer(GLuint FramebufferName)
  {
#ifdef NSIGHT_DEBUGGER
    glBindFramebuffer(GL_FRAMEBUFFER, FramebufferName);
    GLenum Status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
#else
    GLenum Status = glCheckNamedFramebufferStatusEXT(FramebufferName, GL_FRAMEBUFFER);
#endif

    switch(Status)
    {
    case GL_FRAMEBUFFER_UNDEFINED:
      fprintf(stdout, "OpenGL Error(%s)\n", "GL_FRAMEBUFFER_UNDEFINED");
      break;
    case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT:
      fprintf(stdout, "OpenGL Error(%s)\n", "GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT");
      break;
    case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT:
      fprintf(stdout, "OpenGL Error(%s)\n", "GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT");
      break;
    case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER:
      fprintf(stdout, "OpenGL Error(%s)\n", "GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER");
      break;
    case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER:
      fprintf(stdout, "OpenGL Error(%s)\n", "GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER");
      break;
    case GL_FRAMEBUFFER_UNSUPPORTED:
      fprintf(stdout, "OpenGL Error(%s)\n", "GL_FRAMEBUFFER_UNSUPPORTED");
      break;
    case GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE:
      fprintf(stdout, "OpenGL Error(%s)\n", "GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE");
      break;
    case GL_FRAMEBUFFER_INCOMPLETE_LAYER_TARGETS:
      fprintf(stdout, "OpenGL Error(%s)\n", "GL_FRAMEBUFFER_INCOMPLETE_LAYER_TARGETS");
      break;
    }

    return Status != GL_FRAMEBUFFER_COMPLETE;
  }

  bool checkGLVersion(GLint MajorVersionRequire, GLint MinorVersionRequire)
  {
    GLint MajorVersionContext = 0;
    GLint MinorVersionContext = 0;
    glGetIntegerv(GL_MAJOR_VERSION, &MajorVersionContext);
    glGetIntegerv(GL_MINOR_VERSION, &MinorVersionContext);
    return glf::version(MajorVersionContext, MinorVersionContext) 
      >= glf::version(MajorVersionRequire, MinorVersionRequire);
  }

  bool checkExtension(char const * String)
  {
    GLint ExtensionCount = 0;
    glGetIntegerv(GL_NUM_EXTENSIONS, &ExtensionCount);
    for(GLint i = 0; i < ExtensionCount; ++i)
      if(std::string((char const*)glGetStringi(GL_EXTENSIONS, i)) == std::string(String))
        return true;
    return false;
  }

  void APIENTRY debugOutput
    (
    GLenum source,
    GLenum type,
    GLuint id,
    GLenum severity,
    GLsizei length,
    const GLchar* message,
    GLvoid* userParam
    )
  {
    //FILE* f;
    //f = fopen("debug_output.txt","a");
    //if(f)
    {
      char debSource[32], debType[32], debSev[32];
      bool Error(false);

      if(source == GL_DEBUG_SOURCE_API_ARB)
        strcpy(debSource, "OpenGL");
      else if(source == GL_DEBUG_SOURCE_WINDOW_SYSTEM_ARB)
        strcpy(debSource, "Windows");
      else if(source == GL_DEBUG_SOURCE_SHADER_COMPILER_ARB)
        strcpy(debSource, "Shader Compiler");
      else if(source == GL_DEBUG_SOURCE_THIRD_PARTY_ARB)
        strcpy(debSource, "Third Party");
      else if(source == GL_DEBUG_SOURCE_APPLICATION_ARB)
        strcpy(debSource, "Application");
      else if(source == GL_DEBUG_SOURCE_OTHER_ARB)
        strcpy(debSource, "Other");

      if(type == GL_DEBUG_TYPE_ERROR_ARB)
        strcpy(debType, "error");
      else if(type == GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR_ARB)
        strcpy(debType, "deprecated behavior");
      else if(type == GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR_ARB)
        strcpy(debType, "undefined behavior");
      else if(type == GL_DEBUG_TYPE_PORTABILITY_ARB)
        strcpy(debType, "portability");
      else if(type == GL_DEBUG_TYPE_PERFORMANCE_ARB)
        strcpy(debType, "performance");
      else if(type == GL_DEBUG_TYPE_OTHER_ARB)
        strcpy(debType, "message");

      if(severity == GL_DEBUG_SEVERITY_HIGH_ARB)
      {
        strcpy(debSev, "high");
        Error = true;
      }
      else if(severity == GL_DEBUG_SEVERITY_MEDIUM_ARB)
        strcpy(debSev, "medium");
      else if(severity == GL_DEBUG_SEVERITY_LOW_ARB)
        strcpy(debSev, "low");

      fprintf(stderr,"%s: %s(%s) %d: %s\n", debSource, debType, debSev, id, message);
      assert(!Error);
      //fclose(f);
    }
  }
  inline void logImplementationDependentLimit(GLenum Value, std::string const & String)
  {
    GLint Result(0);
    glGetIntegerv(Value, &Result);
    std::string Message(glf::format("%s: %d", String.c_str(), Result));
#		if(!defined(__APPLE__))
    glDebugMessageInsertARB(GL_DEBUG_SOURCE_APPLICATION_ARB, GL_DEBUG_TYPE_OTHER_ARB, 1, GL_DEBUG_SEVERITY_LOW_ARB, GLsizei(Message.size()), Message.c_str());
#		endif
  }

  bool initDebugOutput()
  {
    glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS_ARB);
    glDebugMessageControlARB(GL_DONT_CARE, GL_DONT_CARE, GL_DONT_CARE, 0, NULL, GL_TRUE);
    glDebugMessageCallbackARB(&glf::debugOutput, NULL);

    return true;
  }
}//namespace glf
