#include "textures.h"
#include "fbo.h" 
#include "glo_external.h"
#include "error.hpp"
#include "tricks.h"

#include <iostream>

namespace glc
{
  std::vector<GLuint> bound_buffers_stack;

	DepthFBO::DepthFBO(GLsizei w, GLsizei h, bool dms): 
  Width(w), Height(h), requires_initialization(true), depth_mode_shadow(dms)
	{

	}
	DepthFBO::~DepthFBO()
	{
    if (!requires_initialization)
		  glDeleteFramebuffers(1, &Buffer); 
	}
	void DepthFBO::Bind()
	{
    bool reqinit = requires_initialization;
    if (requires_initialization)
    {
      requires_initialization = false;
      InitTextures();
      InitBuffer();
    }
     
    bool bound  = 
      ! bound_buffers_stack.empty() && 
        bound_buffers_stack.back() == Buffer;

    bound_buffers_stack.push_back(Buffer);     

    if (!bound)
      glBindFramebuffer(GL_FRAMEBUFFER, Buffer);

	  if (reqinit)
      BoundClear();
   
	}
	void DepthFBO::Unbind()
	{
    assert(bound_buffers_stack.back() == Buffer);
    bound_buffers_stack.pop_back();
    if (bound_buffers_stack.empty())
    {
      glBindFramebuffer(GL_FRAMEBUFFER, 0);
      //std::cout << "bound 0 by unbound" << std::endl;
    }
    else
    {
      GLuint id = bound_buffers_stack.back();
      if (id != Buffer)
      {
        glBindFramebuffer(GL_FRAMEBUFFER, id);
      }
    }
	}
	void DepthFBO::SetCommonTextureParam(TextureNameReference& t)
	{
		t.SetParameter(GL_TEXTURE_BASE_LEVEL, 0);
		t.SetParameter(GL_TEXTURE_MAX_LEVEL, 0);
		t.SetParameter(GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		t.SetParameter(GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
		t.SetParameter(GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER );
		t.SetParameter(GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER );
	}
	void DepthFBO::Clear()
	{
    Bind();
    BoundClear();
    Unbind();
	}

	void DepthFBO::InitBuffer()
	{
		glGenFramebuffers(1, &Buffer);
		glNamedFramebufferTexture2DEXT(Buffer, GL_DEPTH_ATTACHMENT, GL_TEXTURE_2D, DepthTexture, 0);    
		InitColorAttachements();
		ChooseBuffersToDraw();
		if(glf::checkFramebuffer(Buffer))
			throw std::runtime_error("frame buffer not complete");

	}

	void DepthFBO::InitTextures()
	{ 
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glPixelStorei(GL_PACK_ALIGNMENT, 1);
		SetCommonTextureParam(DepthTexture);
    if (depth_mode_shadow)
    {
		  DepthTexture.SetParameter(GL_TEXTURE_COMPARE_FUNC, GL_LEQUAL);
		  DepthTexture.SetParameter(GL_TEXTURE_COMPARE_MODE, GL_COMPARE_R_TO_TEXTURE);
    }
    else {
      DepthTexture.SetParameter(GL_TEXTURE_COMPARE_MODE, GL_NONE);
    }
		glTextureStorage2DEXT(DepthTexture, GL_TEXTURE_2D, GLint(1), GL_DEPTH_COMPONENT32, Width, Height);
		//glGenerateTextureMipmapEXT(DepthTexture, GL_TEXTURE_2D);

		glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	}
	void DepthFBO::ChooseBuffersToDraw()
	{
		glFramebufferDrawBufferEXT(Buffer, GL_NONE);
	}
	std::vector<float> DepthFBO::ReadDepth()
	{
    Bind();
		std::vector<float> res(Width*Height);
		glReadPixels(0, 0, Width, Height,GL_DEPTH_COMPONENT, GL_FLOAT, &res[0]);
    Unbind();
		return res;
	}

  void DepthFBO::BoundClear()
  {
    float Depth(1.0f);
    glClearBufferfv(GL_DEPTH , 0, &Depth); 
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);
  }

  std::vector<float> DepthFBO::ReadChannel( int i )
  {
    
    assert(i < 4);
    if (i >= 4)
      throw "bad channel";

    Bind();
    std::vector<float> res(Width*Height);
    GLenum format[] =  { GL_RED, GL_GREEN, GL_BLUE, GL_ALPHA };
    glReadPixels(0, 0, Width, Height, format[i], GL_FLOAT, &res[0]);
    Unbind();
    return res;


  }

	CommonFBO::CommonFBO( GLsizei w, GLsizei h, bool shadow_mode) : DepthFBO(w, h, shadow_mode) { }
	void CommonFBO::InitColorAttachements()
	{
		glNamedFramebufferTexture2DEXT(Buffer, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, ColorTexture, 0);
	}
	void CommonFBO::InitTextures()
	{
		DepthFBO::InitTextures();
		SetCommonTextureParam(ColorTexture);
		glTextureStorage2DEXT(ColorTexture, GL_TEXTURE_2D, GLint(1), GL_RGBA32F, Width, Height);
	}

	void CommonFBO::ChooseBuffersToDraw()
	{
		glFramebufferDrawBufferEXT(Buffer, GL_COLOR_ATTACHMENT0);
	}

	std::vector<unsigned char> CommonFBO::ReadRGB()
	{
    Bind();
		std::vector<unsigned char> res(Width*Height*3);
		glReadPixels(0, 0, Width, Height, GL_RGB, GL_UNSIGNED_BYTE, &res[0]);
    Unbind();
		return res;
	}

  void CommonFBO::BoundClear()
  {
    DepthFBO::BoundClear();
    glClearBufferfv(GL_COLOR, 0, &glm::vec4(0.0f, 0.0f, 0.0f, 0.0f)[0]);
  }

	void Averaging::Enable( int samples )
	{
		int n =samples;
		glEnable(GL_BLEND);
		glBlendColor(1.0 /n, 1.0 /n, 1.0 /n, 1.0 /n);		
    // set incomming collor to be scaled by 1/n 
		glBlendFunc(GL_CONSTANT_COLOR, GL_ONE);

	}

	void Averaging::Disable()
	{
		glDisable(GL_BLEND);
	}
  
  void AlphaBlending::Enable()
  {
    glEnable(GL_BLEND);
    // use alpha from src
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  }

  void AlphaBlending::Disable()
  {
    glDisable(GL_BLEND);
  }

  void MovingAveraging::Enable( float lambda)
  {
    glEnable(GL_BLEND);
    glBlendColor(lambda, lambda, lambda, lambda);		
    // set a convex cominbation of src and dst colors
    glBlendFunc(GL_CONSTANT_COLOR, GL_ONE_MINUS_CONSTANT_COLOR);

  }

  void MovingAveraging::Disable()
  {
    glDisable(GL_BLEND);
  }


}