﻿#include "programs.h"
#include "textures.h"

#include "glo_external.h"
#include "error.hpp"
#include "compiler.hpp"

namespace glc
{
  CommonProgram::~CommonProgram()
  {
    glDeleteProgram(ProgramName);
  }

  CommonProgram::CommonProgram():
  ProgramStages(0)
  {
    ProgramName = glCreateProgram();
    glProgramParameteri(ProgramName, GL_PROGRAM_SEPARABLE, GL_TRUE);
  }

  void CommonProgram::InstallShader( CommonShader& shader )
  { 
    glAttachShader(ProgramName, shader.ShaderName);
    ProgramStages = ProgramStages | shader.ShaderUsageBit;
  }

  void CommonProgram::Link()
  {
    glLinkProgram(ProgramName);
    if (! ::glf::checkProgram(ProgramName))
      throw std::runtime_error("incorrect shader program");  
  }

  chain_reference<CommonProgram, CommonShader&> CommonProgram::operator<<( CommonShader& sh )
  {
    InstallShader(sh);
    return *this;
  }

  GLuint CommonProgram::getUniformLocation( const char* name )
  {
    return glGetUniformLocation(this->ProgramName, name);
  }
  void CommonProgram::setUniform(GLuint loc,  float val )
  {
    glProgramUniform1f(ProgramName, loc, val);
  }
  void CommonProgram::setUniform(GLuint loc, mat4 val )
  {
    glProgramUniformMatrix4fv(ProgramName, loc, 1, GL_FALSE, &val[0][0]);
  }

  void CommonProgram::setUniform( GLuint loc, vec2 val )
  {
    glProgramUniform2f(ProgramName, loc, val.x, val.y);
  }
  void CommonProgram::setUniform( GLuint loc, ivec2 val )
  {
    glProgramUniform2i(ProgramName, loc, val.x, val.y);
  }
  void CommonProgram::setUniform( GLuint loc, vec3 val )
  {
    glProgramUniform3f(ProgramName, loc, val.x, val.y, val.z);
  }

  void CommonProgram::BindTexture( GLuint binding, TextureNameReference tex ) 
  {
    assert(binding, "zero unit is reserved for temporary stuff");
    glBindMultiTextureEXT(GL_TEXTURE0 + binding, GL_TEXTURE_2D, tex);
  }
  GLint CommonProgram::GetActiveAttributesCount()
  {
    GLint res;
    glGetProgramiv(this->ProgramName, GL_ACTIVE_ATTRIBUTES, &res);
    return res; 
  }

  CommonProgram::ResourceInfo CommonProgram::GetActiveAttrib(GLuint index)
  {
    char name[256];
    GLint sz;
    GLenum typ;
    glGetActiveAttrib(this->ProgramName, index, 255, NULL, &sz, &typ, name);
    ResourceInfo a;
    a.size = sz;
    a.name = name;
    a.type = typ;
    a.location = glGetAttribLocation(this->ProgramName, name);
    return a;
  }

  GLint CommonProgram::GetActiveUniformCount()
  {
    GLint res;
    glGetProgramiv(this->ProgramName, GL_ACTIVE_UNIFORMS, &res);
    return res; 
  }

  CommonProgram::ResourceInfo CommonProgram::GetActiveUnifrom( GLuint index )
  {
    char name[1024];
    GLint sz;
    GLenum typ;
     
    ResourceInfo u;
    glGetActiveUniformName( this->ProgramName, index, 1023,  0, name);
    u.name = name;
    glGetActiveUniformsiv(this->ProgramName, 1, &index, GL_UNIFORM_TYPE, (GLint*)&u.type);
    glGetActiveUniformsiv(this->ProgramName, 1, &index, GL_UNIFORM_SIZE, &u.size);
    u.location= glGetUniformLocation(this->ProgramName, name);
    return u;
  }

  GLint CommonProgram::GetResourceProp( GLenum inteface, GLint index, GLenum prop )
  {
    GLint res;
    glGetProgramResourceiv(this->ProgramName, inteface, index, 1, &prop, 1, NULL, &res);
    return res;
  }


  GLint CommonProgram::GetActiveResourceCount( GLenum inteface )
  {
    GLenum res;
    glGetProgramInterfaceiv(this->ProgramName, inteface, GL_ACTIVE_RESOURCES, (GLint*)&res);
    return res;
  }

  std::string CommonProgram::GetResourceName( GLenum interface, GLint index )
  {
    char name[1024];
    glGetProgramResourceName(this->ProgramName, interface, index, 1023, NULL, name);
    return name;
  }

  GLint CommonProgram::getUniformi( std::string name )
  {
    GLint res;
     
    glGetUniformiv(this->ProgramName, getUniformLocation(name.c_str()), &res);
    return res;
  }

  CommonShader::CommonShader( GLuint shader_name, GLenum shader_usage_bit):
  ShaderUsageBit(shader_usage_bit), ShaderName(shader_name)
  {
  }

  PCommonShader CommonShaderCompiler::CreateVertexShader(std::string filename)
  {
    GLuint name = Compiler->create(GL_VERTEX_SHADER, filename, "--version 420 --profile core");
    return PCommonShader(new CommonShader(name, GL_VERTEX_SHADER_BIT));
  }
  PCommonShader CommonShaderCompiler::CreateFragmentShader(std::string filename)
  {
    GLuint name = Compiler->create(GL_FRAGMENT_SHADER, filename, "--version 420 --profile core");
    return PCommonShader(new CommonShader(name, GL_FRAGMENT_SHADER_BIT));
  }

  CommonShaderCompiler::CommonShaderCompiler()
  {
    this->Compiler = std::shared_ptr<glf::compiler>(new glf::compiler());
  }

  void CommonPipeline::InstallProgram(const CommonProgram& p)
  {
    glUseProgramStages(PipelineName, p.ProgramStages, p.ProgramName);
  }

  CommonPipeline::CommonPipeline()
  {
    glGenProgramPipelines(1, &PipelineName);
  }

  CommonPipeline::~CommonPipeline()
  {
    glDeleteProgramPipelines(1, &PipelineName);
  }

  void CommonPipeline::Bind()
  {
    glBindProgramPipeline(PipelineName);
  }

  void CommonPipeline::Unbind()
  {
    glBindProgramPipeline(0);
  }

  VertexFragmentProgram::VertexFragmentProgram(std::string frag_shader, std::string vert_shader ) 
  {
    auto compiler =  make_auto_ptr(new CommonShaderCompiler());
    *this 
      << *compiler->CreateFragmentShader(frag_shader)
      << *compiler->CreateVertexShader(vert_shader);

    this->Link();
  }

};