#include "glo_external.h"
#include <GL/glfw.h>

#include "window.h"

#include "window_callbacks.h"


#include <string.h>
#include <stdexcept>
#include <iostream>



using namespace glf;

namespace glf
{
  enum mouse_button
  {
    MOUSE_BUTTON_NONE = 0,
    MOUSE_BUTTON_LEFT = (1 << 0),
    MOUSE_BUTTON_RIGHT = (1 << 1),
    MOUSE_BUTTON_MIDDLE = (1 << 2)

  };

  bool initDebugOutput();
  bool checkGLVersion(GLint MajorVersionRequire, GLint MinorVersionRequire);
  bool checkExtension(char const * String);
}
#define GLF_CONTEXT_CORE_PROFILE_BIT 0x0001
#define GLF_CONTEXT_COMPATIBILITY_PROFILE_BIT 0x0002
#define GLF_CONTEXT_ES2_PROFILE_BIT 0x0004


bool window::Closing()
{
  return KeyPressed[GLFW_KEY_ESC] == 1;
}

glf::window::window( glm::ivec2 const & Size, string title, GLInfo v) :
Size(Size),
  title(title),
  glinfo(v),
  MouseOrigin(Size >> 1),
  MouseCurrent(Size >> 1),
  TranlationOrigin(0, 0),
  TranlationCurrent(0, 0),
  RotationOrigin(0), 
  RotationCurrent(0),
  MouseButtonFlags(0)
{
  memset(KeyPressed, 0, sizeof(KeyPressed));
}


window* glf::window::Instance()
{
  return pinstance;
}

void glf::window::Open( glm::ivec2 const & size, string title, GLInfo v)
{
  if (!window::pinstance)
  {
    instance = window(size, title, v);
    pinstance = &instance; 
    pinstance->InitWindow(); 
  }
  else
  {
    std::cout << "window already opened";
    throw std::runtime_error("window already opened");
  }
}

glf::window::~window()
{
  glfwTerminate();
}

void glf::window::InitWindow()
{
  glfwInit();
  glfwEnable(GLFW_AUTO_POLL_EVENTS); /* No explicit call to glfwPollEvents() */

  set_window_hint(GLF_CONTEXT_CORE_PROFILE_BIT, glinfo);

  GLboolean Result = glfwOpenWindow(Size.x, Size.y, 8, 8, 8, 32, 0, 0, glinfo.full_screen ? GLFW_FULLSCREEN : GLFW_WINDOW);

  assert(Result == GL_TRUE);

  glfwSetWindowTitle(title.c_str());
  glfwSetMouseButtonCallback(mouse_button_callback);
  glfwSetMousePosCallback(cursor_position_callback);
  glfwSetWindowCloseCallback(close_callback);
  glfwSetKeyCallback(key_callback);

  gl::init();
}

void glf::window::SwapBuffers()
{
  glfwSwapBuffers();
}

void glf::window::EnableDebug()
{

  bool Validated = true;
  Validated = Validated && glf::checkGLVersion(glinfo.major, glinfo.minor);
  assert(Validated);
#ifndef NDEBUG
  if(glf::checkExtension("GL_ARB_debug_output"))
    Validated = glf::initDebugOutput();
  assert(Validated);
#endif

}

window* glf::window::pinstance = 0;


window glf::window::instance;

static glf::window& Window() { return  *window::Instance(); }


void cursor_position_callback(int x, int y)
{
  auto& w = Window();
  w.MouseCurrent = glm::ivec2(x, y);
  w.TranlationCurrent = w.MouseButtonFlags & glf::MOUSE_BUTTON_LEFT ? w.TranlationOrigin + (w.MouseCurrent - w.MouseOrigin) / 10.f : w.TranlationOrigin;
  w.RotationCurrent = w.MouseButtonFlags & glf::MOUSE_BUTTON_RIGHT ? w.RotationOrigin + (w.MouseCurrent - w.MouseOrigin) : w.RotationOrigin;
}

void mouse_button_callback(int Button, int Action)
{
  switch(Action)
  {
  case GLFW_PRESS:
    {
      Window().MouseOrigin = Window().MouseCurrent;
      switch(Button)
      {
      case GLFW_MOUSE_BUTTON_LEFT:
        {
          Window().MouseButtonFlags |= glf::MOUSE_BUTTON_LEFT;
          Window().TranlationOrigin = Window().TranlationCurrent;
        }
        break;
      case GLFW_MOUSE_BUTTON_MIDDLE:
        {
          Window().MouseButtonFlags |= glf::MOUSE_BUTTON_MIDDLE;
        }
        break;
      case GLFW_MOUSE_BUTTON_RIGHT:
        {
          Window().MouseButtonFlags |= glf::MOUSE_BUTTON_RIGHT;
          Window().RotationOrigin = Window().RotationCurrent;
        }
        break;
      }
    }
    break;
  case GLFW_RELEASE:
    {
      switch(Button)
      {
      case GLFW_MOUSE_BUTTON_LEFT:
        {
          Window().TranlationOrigin += (Window().MouseCurrent - Window().MouseOrigin) / 10.f;
          Window().MouseButtonFlags &= ~glf::MOUSE_BUTTON_LEFT;
        }
        break;
      case GLFW_MOUSE_BUTTON_MIDDLE:
        {
          Window().MouseButtonFlags &= ~glf::MOUSE_BUTTON_MIDDLE;
        }
        break;
      case GLFW_MOUSE_BUTTON_RIGHT:
        {
          Window().RotationOrigin += Window().MouseCurrent - Window().MouseOrigin;
          Window().MouseButtonFlags &= ~glf::MOUSE_BUTTON_RIGHT;
        }
        break;
      }
    }
    break;
  }
}

void GLFWCALL key_callback(int key, int action)
{
  switch(action)
  {
  case GLFW_PRESS:
    {
      Window().KeyPressed[key] = 1;
      break;
    }
  case GLFW_RELEASE:
    {
      Window().KeyPressed[key] = 0;
      break;
    }
  }
}

int GLFWCALL close_callback()
{
  Window().KeyPressed[GLFW_KEY_ESC] = 1;
  return GL_FALSE;
}


#include "error.hpp"
void set_window_hint(int Profile, const GLInfo & gl) 
{
  glfwOpenWindowHint(GLFW_WINDOW_NO_RESIZE, gl.full_screen ? GL_FALSE : GL_TRUE);
  assert(version(gl.major, gl.minor) >= version(3, 2));

  glfwOpenWindowHint(GLFW_OPENGL_VERSION_MAJOR, gl.major);
  glfwOpenWindowHint(GLFW_OPENGL_VERSION_MINOR, gl.minor);

#			if defined(__APPLE__)
  glfwOpenWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
#			elif defined(__INTEL__)
  glfwOpenWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_COMPAT_PROFILE);
#			else
  glfwOpenWindowHint(GLFW_OPENGL_PROFILE, Profile == GLF_CONTEXT_CORE_PROFILE_BIT ? GLFW_OPENGL_CORE_PROFILE : GLFW_OPENGL_COMPAT_PROFILE);
  glfwOpenWindowHint(GLFW_OPENGL_FORWARD_COMPAT, Profile == GLF_CONTEXT_CORE_PROFILE_BIT ? GL_TRUE : GL_FALSE);
#			endif

#			if defined(NDEBUG)
  glfwOpenWindowHint(GLFW_OPENGL_DEBUG_CONTEXT, GL_FALSE);
#			else
  glfwOpenWindowHint(GLFW_OPENGL_DEBUG_CONTEXT, GL_TRUE);
#			endif
}