#include "public_modules.h"
#include <functional>
#include <memory>

#include <scene-core/sra-api/sra_scene.h>
#include <scene-core/ErrorAcelValue.h>
 
#include "DummyAcel.h"
#include "GenericConverter.h"
#include "StreamConverters.h"
#include "MeshesConverter.h"

#include "CameraModule.h"
#include "MeshRenderingModule.h"
#include "DepthRenderingModule.h"
#include "SplashDrawModule.h"
#include "ShadowComputationModule.h"

using namespace sra_render;

using std::shared_ptr;
typedef shared_ptr<sra::IAcel> PAcel;
typedef shared_ptr<sra::IConverter> PConverter;


PAcel ModViewPoint::Execute()
{
  auto acel = new DummyAcel<glm::mat4>(module->GetCurrentMVP());
  return PAcel(acel);
}

ModViewPoint::ModViewPoint(shared_ptr<CameraModule> c ): module(c)
{
  CameraSettings = 
    make_auto_ptr(new GenericConverter<CameraAcelAcessor>(
    [](PAcel f){ return CameraAcelAcessor(f); },
    [this](CameraAcelAcessor c){ module->AssignCameraParameters(c); }));
}


template <typename TArg >
PConverter make_converter_from_dummy(std::function<void( TArg)> assigner)
{
  return PConverter(new GenericConverter<TArg>(
    [](PAcel a){ return extract_acel_value<TArg>(a); },
    assigner));
}

ModRenderMesh::ModRenderMesh(shared_ptr<MeshRenderingModule> mod): module(mod)
{
  auto images = std::make_shared<ImageStreamConverter>();
  Texture = PConverter(new GenericConverter<PImageTexture>(
    [=](PAcel f){ return images->LoadTexture(f); }, 
    [this](PImageTexture t){module->AssignTextureDiffuse(t); }
  ));
  ShadowMap = make_converter_from_dummy<TextureNameReference>(
    [this](TextureNameReference t)->void {module->BindTextureShadowMap(t); } );
  ShadowTransform = make_converter_from_dummy<glm::mat4>( [this](glm::mat4 t)->void {module->AssignWorldToLight(t); } );

  Transform = make_converter_from_dummy<glm::mat4>( [this](glm::mat4 t)->void {module->SetMVP(t); } );

  Mesh = make_auto_ptr(new MeshesConverter(module));
}

ModRenderDepthMap::ModRenderDepthMap( shared_ptr<DepthMapRenderingModule> m):module(m)
{
  auto images = std::make_shared<ImageStreamConverter>();
  auto depth = std::make_shared<DepthStreamConverter>();
  
  DepthMap = PConverter(new GenericConverter<PImageTexture>(
    [=](PAcel f){ return depth->LoadTexture(f); }, 
    [this](PImageTexture t){module->AssignTextureDepth(t); }
  ));

  Texture = PConverter(new GenericConverter<PImageTexture>(
    [=](PAcel f){ return images->LoadTexture(f); }, 
    [this](PImageTexture t){module->AssignTextureDiffuse(t); }
  ));

  Transform = 
    PConverter(new GenericConverter<glm::mat4>(
    [](PAcel f){ return extract_acel_value<glm::mat4>(f); },
    [this](glm::mat4 t){module->SetMVP(t); }
  ));

  ShadowMap = make_converter_from_dummy<TextureNameReference>(
    [this](TextureNameReference t)->void {module->BindTextureShadowMap(t); } );
  ShadowTransform = make_converter_from_dummy<glm::mat4>( [this](glm::mat4 t)->void {module->AssignWorldToLight(t); } );

  CameraSettings =  
    PConverter(new GenericConverter<CameraAcelAcessor>(
    [](PAcel f){  return CameraAcelAcessor(f); },
    [this](CameraAcelAcessor c) { module->AssignCameraParameters(c); } 
  ));
}

sra_render::ModDrawSplash::ModDrawSplash( std::shared_ptr<SplashDrawModule> m ): module(m)
{
  auto images = std::make_shared<ImageStreamConverter>();
  Texture = PConverter(new GenericConverter<PImageTexture>(
    [=](PAcel f){ return images->LoadTexture(f); }, 
    [this](PImageTexture t){module->AssignTextureDiffuse(t); }
  ));
}

std::shared_ptr<sra::IAcel> sra_render::ModDrawSplash::Execute()
{
  module->Draw();
  return std::make_shared<ErrorAcel>();

}


PAcel sra_render::ModRenderMesh::Execute()
{
  module->Draw();
  return std::make_shared<ErrorAcel>();
}

PAcel sra_render::ModRenderDepthMap::Execute()
{
  module->Draw();
  return std::make_shared<ErrorAcel>();
}



void sra_render::ModComputeShadows::Begin() { module->Begin(); }

void sra_render::ModComputeShadows::End() { module->End(); }

std::shared_ptr<sra::IAcel> sra_render::ModComputeShadows::Execute()
{
  auto tex = module->GetShadowMap();
  auto acel = new DummyAcel<glc::TextureNameReference>(tex);
  return PAcel(acel);
}
sra_render::ModComputeShadows::ModComputeShadows(std::shared_ptr<ShadowComputationModule> m) :module(m)
{
}

